@@ -14,11 +14,13 @@ from kivy.properties import ObjectProperty, ListProperty, BooleanProperty, Bound
     NumericProperty
 from kivy.uix.widget import Widget
 from kivy.vector import Vector
+from matplotlib.colors import to_hex
 from matplotlib.backends.backend_agg import FigureCanvasAgg
 from matplotlib import cbook
 from weakref import WeakKeyDictionary
 from kivy.metrics import dp
 import numpy as np
+from kivy.utils import get_color_from_hex
 
 class MatplotFigureScatter(Widget):
     """Widget to show a matplotlib figure in kivy.
@@ -47,13 +49,19 @@ class MatplotFigureScatter(Widget):
     pos_y_rect_ver=NumericProperty(0)   
     invert_rect_ver = BooleanProperty(False)
     invert_rect_hor = BooleanProperty(False)
-    legend_instance = ObjectProperty(None)
+    legend_instance = ObjectProperty(None, allownone=True)
     legend_do_scroll_x = BooleanProperty(True)
     legend_do_scroll_y = BooleanProperty(True)    
     do_pan_x = BooleanProperty(True)
     do_pan_y = BooleanProperty(True)    
     do_zoom_x = BooleanProperty(True)
-    do_zoom_y = BooleanProperty(True)  
+    do_zoom_y = BooleanProperty(True)
+    fast_draw = BooleanProperty(True) #True will don't draw axis
+    xsorted = BooleanProperty(False) #to manage x sorted data
+    minzoom = NumericProperty(dp(40))
+    multi_xdata = BooleanProperty(False)
+    multi_xdata_res = NumericProperty(dp(20))
+    hover_instance = ObjectProperty(None, allownone=True)
     
     def on_figure(self, obj, value):
         self.figcanvas = _FigureCanvas(self.figure, self)
@@ -69,7 +77,8 @@ class MatplotFigureScatter(Widget):
             ax=self.figure.axes[0]
             patch_cpy=copy.copy(ax.patch)
             patch_cpy.set_visible(False)
-            ax.spines[:].set_zorder(10)
+            for pos in ['right', 'top', 'bottom', 'left']:
+                ax.spines[pos].set_zorder(10)
             patch_cpy.set_zorder(9)
             self.background_patch_copy= ax.add_patch(patch_cpy)
 
@@ -79,10 +88,17 @@ class MatplotFigureScatter(Widget):
             #set default xmin/xmax and ymin/ymax
             self.xmin,self.xmax = self.axes.get_xlim()
             self.ymin,self.ymax = self.axes.get_ylim()
+
+        if self.legend_instance:
+            self.legend_instance.reset_legend()
+            self.legend_instance=None
             
         # Texture
         self._img_texture = Texture.create(size=(w, h))
 
+        #close last figure in memory (avoid max figure warning)
+        matplotlib.pyplot.close()
+
     def __init__(self, **kwargs):
         super(MatplotFigureScatter, self).__init__(**kwargs)
         
@@ -93,17 +109,14 @@ class MatplotFigureScatter(Widget):
         self.xmax = None
         self.ymin = None
         self.ymax = None
+        self.lines = []
+        self.scatters =[]
         
         #option
-        self.zoompan = None
-        self.fast_draw=True
-        self.draw_left_spline=False #available only when fast_draw is True
         self.touch_mode='pan'
         self.hover_on = False
-        self.xsorted = False #to manage x sorted data
-        self.minzoom = dp(40) #minimum pixel distance to apply zoom
-        self.multi_xdata=False
-        self.multi_xdata_res = dp(20)
+        self.cursor_xaxis_formatter=None #used matplotlib formatter to display x cursor value
+        self.cursor_yaxis_formatter=None #used matplotlib formatter to display y cursor value
 
         #zoom box coordonnate
         self.x0_box = None
@@ -127,6 +140,10 @@ class MatplotFigureScatter(Widget):
         self.anchor_x = None
         self.anchor_y = None 
 
+        #manage hover data
+        self.x_hover_data = None
+        self.y_hover_data = None
+
         #manage back and next event
         self._nav_stack = cbook.Stack()
         self.set_history_buttons()   
@@ -350,20 +367,76 @@ class MatplotFigureScatter(Widget):
                         y = self.y_cursor[good_index[idx_best]] 
                     ax=line.axes                     
                     
-                self.set_cross_hair_visible(True)
+                if not self.hover_instance:
+                    self.set_cross_hair_visible(True)
                 
                 # update the cursor x,y data
                 self.horizontal_line.set_ydata(y)
                 self.vertical_line.set_xdata(x)
 
                 #x y label
-                if self.scatter_label  and idx_best > len(good_line)-1: 
-                    if self.multi_xdata:                                
-                        self.text.set_text(f"{self.scatter_label[good_index2_scatter[idx_best]]} x={x}, y={y}")
+                if self.hover_instance:                     
+                    xy_pos = ax.transData.transform([(x,y)]) 
+                    self.x_hover_data = x
+                    self.y_hover_data = y
+                    self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                    self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+                    self.hover_instance.show_cursor=True
+                        
+                    if self.cursor_xaxis_formatter:
+                        x = self.cursor_xaxis_formatter.format_data(x)
+                    if self.cursor_yaxis_formatter:
+                        y = self.cursor_yaxis_formatter.format_data(y) 
+                    self.hover_instance.label_x_value=f"{x}"
+                    self.hover_instance.label_y_value=f"{y}"
+            
+                    self.hover_instance.ymin_line = float(ax.bbox.bounds[1])  + self.y
+                    self.hover_instance.ymax_line = float(ax.bbox.bounds[1] + ax.bbox.bounds[3])  + self.y
+
+                    if self.scatter_label  and idx_best > len(good_line)-1: 
+                        if self.multi_xdata:                                
+                            self.hover_instance.custom_label = self.scatter_label[good_index2_scatter[idx_best]]
+                        else:                   
+                            self.hover_instance.custom_label = self.scatter_label[good_index_scatter[idx_best]]
+                    if line:
+                        self.hover_instance.custom_color = get_color_from_hex(to_hex(line.get_color()))
+                    elif scatter:
+                        if self.multi_xdata:
+                            if len(self.x_cursor)==len(scatter.get_facecolors()):
+                                color=scatter.get_facecolors()[good_index2_scatter[idx_best]]
+                            else:
+                                color = scatter.get_facecolors()
+                            self.hover_instance.custom_color = get_color_from_hex(to_hex(color))
+                        else:
+                            if len(self.x_cursor)==len(scatter.get_facecolors()):
+                                color=scatter.get_facecolors()[good_index_scatter[idx_best]]
+                            else:
+                                color = scatter.get_facecolors()                           
+                            self.hover_instance.custom_color = get_color_from_hex(to_hex(color))
+                                                                        
+                    if self.hover_instance.x_hover_pos>self.x+self.axes.bbox.bounds[2] + self.axes.bbox.bounds[0] or \
+                        self.hover_instance.x_hover_pos<self.x+self.axes.bbox.bounds[0] or \
+                        self.hover_instance.y_hover_pos>self.y+self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] or \
+                        self.hover_instance.y_hover_pos<self.y+self.axes.bbox.bounds[1]:               
+                        self.hover_instance.hover_outside_bound=True
                     else:
-                        self.text.set_text(f"{self.scatter_label[good_index_scatter[idx_best]]} x={x}, y={y}")
-                else:
-                    self.text.set_text(f"x={x}, y={y}")
+                        self.hover_instance.hover_outside_bound=False                      
+                    
+                    return
+                else:                
+
+                    if self.cursor_xaxis_formatter:
+                        x = self.cursor_xaxis_formatter.format_data(x)
+                    if self.cursor_yaxis_formatter:
+                        y = self.cursor_yaxis_formatter.format_data(y) 
+                        
+                    if self.scatter_label  and idx_best > len(good_line)-1: 
+                        if self.multi_xdata:                                
+                            self.text.set_text(f"{self.scatter_label[good_index2_scatter[idx_best]]} x={x}, y={y}")
+                        else:
+                            self.text.set_text(f"{self.scatter_label[good_index_scatter[idx_best]]} x={x}, y={y}")
+                    else:
+                        self.text.set_text(f"x={x}, y={y}")
 
                 #blit method (always use because same visual effect as draw)
                 if self.background is None:
@@ -385,7 +458,13 @@ class MatplotFigureScatter(Widget):
 
             #if touch is too far, hide cross hair cursor
             else:
-                self.set_cross_hair_visible(False)  
+                self.set_cross_hair_visible(False) 
+                if self.hover_instance:
+                    self.hover_instance.x_hover_pos=self.x
+                    self.hover_instance.y_hover_pos=self.y      
+                    self.hover_instance.show_cursor=False
+                    self.x_hover_data = None
+                    self.y_hover_data = None                
 
     def home(self) -> None:
         """ reset data axis
@@ -517,6 +596,26 @@ class MatplotFigureScatter(Widget):
             bytes(self._bitmap), colorfmt="rgba", bufferfmt='ubyte')
         self._img_texture.flip_vertical()
 
+        if self.hover_instance:
+            #update hover pos if needed
+            if self.hover_instance.show_cursor and self.x_hover_data and self.y_hover_data:        
+                xy_pos = self.axes.transData.transform([(self.x_hover_data,self.y_hover_data)]) 
+                self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+     
+                # ymin,ymax=self.axes.get_ylim()
+                # ylim_pos = self.axes.transData.transform([(ymin,ymax)])
+                self.hover_instance.ymin_line = float(self.axes.bbox.bounds[1]) + self.y
+                self.hover_instance.ymax_line = float(self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] )+ self.y
+    
+                if self.hover_instance.x_hover_pos>self.x+self.axes.bbox.bounds[2] + self.axes.bbox.bounds[0] or \
+                    self.hover_instance.x_hover_pos<self.x+self.axes.bbox.bounds[0] or \
+                    self.hover_instance.y_hover_pos>self.y+self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] or \
+                    self.hover_instance.y_hover_pos<self.y+self.axes.bbox.bounds[1]:               
+                    self.hover_instance.hover_outside_bound=True
+                else:
+                    self.hover_instance.hover_outside_bound=False
+
     def transform_with_touch(self, event):
         """ manage touch behaviour. based on kivy scatter method"""
         # just do a simple one finger drag
@@ -593,11 +692,31 @@ class MatplotFigureScatter(Widget):
             changed = True
         return changed
 
+    def on_motion(self,*args):
+        '''Kivy Event to trigger mouse event on motion
+           `enter_notify_event`.
+        '''
+        pos = args[1]
+        newcoord = self.to_widget(pos[0], pos[1])
+        x = newcoord[0]
+        y = newcoord[1]
+        inside = self.collide_point(x,y)
+        if inside: 
+
+            # will receive all motion events.
+            if self.figcanvas and self.hover_instance:
+                #avoid in motion if touch is detected
+                if not len(self._touches)==0:
+                    return
+                FakeEventScatter.x=x
+                FakeEventScatter.y=y
+                self.hover(FakeEventScatter)
+                
     def on_touch_down(self, event):
         """ Manage Mouse/touch press """
         x, y = event.x, event.y
 
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
             if self.legend_instance:
                 if self.legend_instance.box.collide_point(x, y):
                     if self.touch_mode!='drag_legend':
@@ -702,7 +821,7 @@ class MatplotFigureScatter(Widget):
                 return True
             
         # stop propagating if its within our bounds
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
 
             if self.do_update:
                 self.update_lim()            
@@ -942,7 +1061,9 @@ class MatplotFigureScatter(Widget):
         self.figcanvas.draw()
         if self.legend_instance:
             self.legend_instance.update_size()        
-
+        if self.hover_instance:
+            self.hover_instance.figwidth = self.width
+            
     def update_lim(self):
         """ update axis lim if zoombox is used"""
         ax=self.axes
@@ -1041,7 +1162,8 @@ class MatplotFigureScatter(Widget):
             x0_max = self.axes.transData.transform([(xmax,ymin)])
             x1=x0_max[0][0]+pos_x             
 
-            self._alpha_ver=1 
+            self._alpha_ver=1
+            self._alpha_hor=0
              
         elif abs(y1-y0)<dp(20) and abs(x1-x0)>self.minzoom:
             self.pos_x_rect_hor=x0
@@ -1054,6 +1176,7 @@ class MatplotFigureScatter(Widget):
             y1=y0_max[0][1]+pos_y         
 
             self._alpha_hor=1
+            self._alpha_ver=0
                         
         else:
             self._alpha_hor=0   
@@ -1102,6 +1225,10 @@ class _FigureCanvas(FigureCanvasAgg):
         self.widget.bt_w = w
         self.widget.bt_h = h
         self.widget._draw_bitmap()
+        
+class FakeEventScatter:
+    x:None
+    y:None
 
 from kivy.factory import Factory
 
