@@ -15,10 +15,12 @@ from kivy.properties import ObjectProperty, ListProperty, BooleanProperty, Bound
 from kivy.uix.widget import Widget
 from kivy.vector import Vector
 from matplotlib.backends.backend_agg import FigureCanvasAgg
+from matplotlib.colors import to_hex
 from matplotlib import cbook
 from weakref import WeakKeyDictionary
 from kivy.metrics import dp
 import numpy as np
+from kivy.utils import get_color_from_hex
 
 class MatplotFigureTwinx(Widget):
     """Widget to show a matplotlib figure in kivy.
@@ -47,13 +49,18 @@ class MatplotFigureTwinx(Widget):
     pos_y_rect_ver=NumericProperty(0)  
     invert_rect_ver = BooleanProperty(False)
     invert_rect_hor = BooleanProperty(False)
-    legend_instance = ObjectProperty(None)
+    legend_instance = ObjectProperty(None, allownone=True)
     legend_do_scroll_x = BooleanProperty(True)
     legend_do_scroll_y = BooleanProperty(True)
     do_pan_x = BooleanProperty(True)
     do_pan_y = BooleanProperty(True)    
     do_zoom_x = BooleanProperty(True)
-    do_zoom_y = BooleanProperty(True)    
+    do_zoom_y = BooleanProperty(True)
+    fast_draw = BooleanProperty(True) #True will don't draw axis
+    xsorted = BooleanProperty(False) #to manage x sorted data
+    minzoom = NumericProperty(dp(40))  
+    twinx = BooleanProperty(False)
+    hover_instance = ObjectProperty(None, allownone=True)
     
     def on_figure(self, obj, value):
         self.figcanvas = _FigureCanvas(self.figure, self)
@@ -69,7 +76,8 @@ class MatplotFigureTwinx(Widget):
             ax=self.figure.axes[0]
             patch_cpy=copy.copy(ax.patch)
             patch_cpy.set_visible(False)
-            ax.spines[:].set_zorder(10)
+            for pos in ['right', 'top', 'bottom', 'left']:
+                ax.spines[pos].set_zorder(10)
             patch_cpy.set_zorder(9)
             self.background_patch_copy= ax.add_patch(patch_cpy)
             
@@ -82,7 +90,8 @@ class MatplotFigureTwinx(Widget):
                 ax2=self.figure.axes[1]
                 patch_cpy_ax2=copy.copy(ax2.patch)
                 patch_cpy_ax2.set_visible(False)
-                ax2.spines[:].set_zorder(10)
+                for pos in ['right', 'top', 'bottom', 'left']:
+                    ax2.spines[pos].set_zorder(10)
                 patch_cpy_ax2.set_zorder(9)
                 self.background_ax2_patch_copy= ax2.add_patch(patch_cpy_ax2) 
                 self.ymin2,self.ymax2 = ax.get_ylim()
@@ -90,9 +99,17 @@ class MatplotFigureTwinx(Widget):
                 self.twinx=False                            
                 self.background_ax2_patch_copy= None 
                 self.ymin2 = None
-                self.ymax2 = None         
+                self.ymax2 = None  
+                
+        if self.legend_instance:
+            self.legend_instance.reset_legend()
+            self.legend_instance=None
+            
         # Texture
         self._img_texture = Texture.create(size=(w, h))
+        
+        #close last figure in memory (avoid max figure warning)
+        matplotlib.pyplot.close()
 
     def __init__(self, **kwargs):
         super(MatplotFigureTwinx, self).__init__(**kwargs)
@@ -106,15 +123,14 @@ class MatplotFigureTwinx(Widget):
         self.ymax = None
         self.ymin2 = None
         self.ymax2 = None
+        self.lines = []
         
         #option
-        self.zoompan = None
-        self.fast_draw=True
-        self.draw_left_spline=False #available only when fast_draw is True
         self.touch_mode='pan'
         self.hover_on = False
-        self.xsorted = False #to manage x sorted data
-        self.minzoom = dp(40) #minimum pixel distance to apply zoom
+        self.cursor_xaxis_formatter=None #used matplotlib formatter to display x cursor value
+        self.cursor_yaxis_formatter=None #used matplotlib formatter to display y cursor value (left axis)
+        self.cursor_yaxis2_formatter=None #used matplotlib formatter to display y cursor value (right axis)       
 
         #zoom box coordonnate
         self.x0_box = None
@@ -138,6 +154,10 @@ class MatplotFigureTwinx(Widget):
         self.anchor_x = None
         self.anchor_y = None 
 
+        #manage hover data
+        self.x_hover_data = None
+        self.y_hover_data = None
+
         #manage back and next event
         self._nav_stack = cbook.Stack()
         self.set_history_buttons()              
@@ -265,8 +285,10 @@ class MatplotFigureTwinx(Widget):
                 line=good_line[idx_best]
                 self.x_cursor, self.y_cursor = line.get_data()
                 x = self.x_cursor[good_index[idx_best]]
-                y = self.y_cursor[good_index[idx_best]]                
-                self.set_cross_hair_visible(True)
+                y = self.y_cursor[good_index[idx_best]] 
+                
+                if not self.hover_instance:
+                    self.set_cross_hair_visible(True)
                 
                 # update the cursor x,y data               
                 ax=line.axes
@@ -278,15 +300,66 @@ class MatplotFigureTwinx(Widget):
                         ratio = (cur_ylim2[1] - cur_ylim2[0]) / (cur_ylim[1] - cur_ylim[0])
                         new_y = (y-cur_ylim2[0])/ ratio + cur_ylim[0]                
                         self.horizontal_line.set_ydata(new_y)
+                        if self.cursor_yaxis2_formatter and not self.hover_instance:
+                            y = self.cursor_yaxis2_formatter.format_data(y) 
                     else:
                         self.horizontal_line.set_ydata(y) 
+                        if self.cursor_yaxis_formatter and not self.hover_instance:
+                            y = self.cursor_yaxis_formatter.format_data(y) 
+                    
                 else:
-                    self.horizontal_line.set_ydata(y)                        
-                self.horizontal_line.set_ydata(y)
+                    self.horizontal_line.set_ydata(y)  
+                    if self.cursor_yaxis_formatter and not self.hover_instance:
+                        y = self.cursor_yaxis_formatter.format_data(y)                       
+                
                 self.vertical_line.set_xdata(x)
-
+                if self.cursor_xaxis_formatter and not self.hover_instance:
+                    x = self.cursor_xaxis_formatter.format_data(x)
+                    
                 #x y label
-                self.text.set_text(f"x={x}, y={y}")
+                if self.hover_instance:                     
+                    xy_pos = ax.transData.transform([(x,y)]) 
+                    self.x_hover_data = x
+                    self.y_hover_data = y
+                    self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                    self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+                    self.hover_instance.show_cursor=True
+                        
+                    if self.twinx:
+                        if ax==self.figure.axes[1]:
+                            if self.cursor_yaxis2_formatter:
+                                y = self.cursor_yaxis2_formatter.format_data(y)                         
+                        else:
+                            if self.cursor_yaxis_formatter:
+                                y = self.cursor_yaxis_formatter.format_data(y) 
+
+                    else: 
+                        if self.cursor_yaxis_formatter:
+                            y = self.cursor_yaxis_formatter.format_data(y) 
+                            
+                    if self.cursor_xaxis_formatter:
+                        x = self.cursor_xaxis_formatter.format_data(x)                            
+                            
+                    self.hover_instance.label_x_value=f"{x}"
+                    self.hover_instance.label_y_value=f"{y}"
+            
+                    self.hover_instance.ymin_line = float(ax.bbox.bounds[1])  + self.y
+                    self.hover_instance.ymax_line = float(ax.bbox.bounds[1] + ax.bbox.bounds[3])  + self.y
+                    
+                    self.hover_instance.custom_label = line.get_label()
+                    self.hover_instance.custom_color = get_color_from_hex(to_hex(line.get_color()))
+                    
+                    if self.hover_instance.x_hover_pos>self.x+self.figure.axes[0].bbox.bounds[2] + self.figure.axes[0].bbox.bounds[0] or \
+                        self.hover_instance.x_hover_pos<self.x+self.figure.axes[0].bbox.bounds[0] or \
+                        self.hover_instance.y_hover_pos>self.y+self.figure.axes[0].bbox.bounds[1] + self.figure.axes[0].bbox.bounds[3] or \
+                        self.hover_instance.y_hover_pos<self.y+self.figure.axes[0].bbox.bounds[1]:               
+                        self.hover_instance.hover_outside_bound=True
+                    else:
+                        self.hover_instance.hover_outside_bound=False                      
+                    
+                    return
+                else:
+                    self.text.set_text(f"x={x}, y={y}")                
 
                 #blit method (always use because same visual effect as draw)                  
                 if self.background is None:
@@ -308,7 +381,13 @@ class MatplotFigureTwinx(Widget):
 
             #if touch is too far, hide cross hair cursor
             else:
-                self.set_cross_hair_visible(False)  
+                self.set_cross_hair_visible(False) 
+                if self.hover_instance:
+                    self.hover_instance.x_hover_pos=self.x
+                    self.hover_instance.y_hover_pos=self.y      
+                    self.hover_instance.show_cursor=False
+                    self.x_hover_data = None
+                    self.y_hover_data = None                
 
     def home(self) -> None:
         """ reset data axis
@@ -437,6 +516,26 @@ class MatplotFigureTwinx(Widget):
         self._img_texture.blit_buffer(
             bytes(self._bitmap), colorfmt="rgba", bufferfmt='ubyte')
         self._img_texture.flip_vertical()
+        
+        if self.hover_instance:
+            #update hover pos if needed
+            if self.hover_instance.show_cursor and self.x_hover_data and self.y_hover_data:        
+                xy_pos = self.axes.transData.transform([(self.x_hover_data,self.y_hover_data)]) 
+                self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+     
+                # ymin,ymax=self.axes.get_ylim()
+                # ylim_pos = self.axes.transData.transform([(ymin,ymax)])
+                self.hover_instance.ymin_line = float(self.figure.axes[0].bbox.bounds[1]) + self.y
+                self.hover_instance.ymax_line = float(self.figure.axes[0].bbox.bounds[1] + self.figure.axes[0].bbox.bounds[3] )+ self.y
+    
+                if self.hover_instance.x_hover_pos>self.x+self.figure.axes[0].bbox.bounds[2] + self.figure.axes[0].bbox.bounds[0] or \
+                    self.hover_instance.x_hover_pos<self.x+self.figure.axes[0].bbox.bounds[0] or \
+                    self.hover_instance.y_hover_pos>self.y+self.figure.axes[0].bbox.bounds[1] + self.figure.axes[0].bbox.bounds[3] or \
+                    self.hover_instance.y_hover_pos<self.y+self.figure.axes[0].bbox.bounds[1]:               
+                    self.hover_instance.hover_outside_bound=True
+                else:
+                    self.hover_instance.hover_outside_bound=False          
 
     def transform_with_touch(self, event):
         """ manage touch behaviour. based on kivy scatter method"""
@@ -524,11 +623,31 @@ class MatplotFigureTwinx(Widget):
             changed = True
         return changed
 
+    def on_motion(self,*args):
+        '''Kivy Event to trigger mouse event on motion
+           `enter_notify_event`.
+        '''
+        pos = args[1]
+        newcoord = self.to_widget(pos[0], pos[1])
+        x = newcoord[0]
+        y = newcoord[1]
+        inside = self.collide_point(x,y)
+        if inside: 
+
+            # will receive all motion events.
+            if self.figcanvas and self.hover_instance:
+                #avoid in motion if touch is detected
+                if not len(self._touches)==0:
+                    return
+                FakeEventTwinx.x=x
+                FakeEventTwinx.y=y
+                self.hover(FakeEventTwinx)
+                
     def on_touch_down(self, event):
         """ Manage Mouse/touch press """
         x, y = event.x, event.y
 
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
             if self.legend_instance:
                 if self.legend_instance.box.collide_point(x, y):
                     if self.touch_mode!='drag_legend':
@@ -645,7 +764,7 @@ class MatplotFigureTwinx(Widget):
                 return True
             
         # stop propagating if its within our bounds
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
 
             if self.do_update:
                 self.update_lim()            
@@ -1074,7 +1193,9 @@ class MatplotFigureTwinx(Widget):
         self.figcanvas.resize_event()
         self.figcanvas.draw() 
         if self.legend_instance:
-            self.legend_instance.update_size()        
+            self.legend_instance.update_size()  
+        if self.hover_instance:
+            self.hover_instance.figwidth = self.width            
 
     def update_lim(self):
         """ update axis lim if zoombox is used"""
@@ -1186,6 +1307,7 @@ class MatplotFigureTwinx(Widget):
             x1=x0_max[0][0]+pos_x
 
             self._alpha_ver=1
+            self._alpha_hor=0
                 
         elif abs(y1-y0)<dp(20) and abs(x1-x0)>self.minzoom:
             self.pos_x_rect_hor=x0
@@ -1198,6 +1320,7 @@ class MatplotFigureTwinx(Widget):
             y1=y0_max[0][1]+pos_y         
 
             self._alpha_hor=1
+            self._alpha_ver=0
                         
         else:
             self._alpha_hor=0   
@@ -1247,6 +1370,10 @@ class _FigureCanvas(FigureCanvasAgg):
         self.widget.bt_h = h
         self.widget._draw_bitmap()
 
+class FakeEventTwinx:
+    x:None
+    y:None
+
 from kivy.factory import Factory
 
 Factory.register('MatplotFigure', MatplotFigureTwinx)
